Model status
Optimal

# Primal solution values
Feasible
Objective 1201500
# Columns 18
STM1 0
ANM1 6
UE1 0
STM2 60
ANM2 6
UE2 0
STM3 60
ANM3 16
UE3 0
STM4 70
ANM4 7
UE4 0
STM5 70
ANM5 12
UE5 0
STM6 75
ANM6 0
